/*
 * $XConsortium: SSMailboxP.h,v 1.16 89/05/11 01:05:56 kit Exp $
 *
 * Copyright 1988 Massachusetts Institute of Technology
 *
 * Permission to use, copy, modify, and distribute this software and its
 * documentation for any purpose and without fee is hereby granted, provided
 * that the above copyright notice appear in all copies and that both that
 * copyright notice and this permission notice appear in supporting
 * documentation, and that the name of M.I.T. not be used in advertising or
 * publicity pertaining to distribution of the software without specific,
 * written prior permission.  M.I.T. makes no representations about the
 * suitability of this software for any purpose.  It is provided "as is"
 * without express or implied warranty.
 *
 * Author:  Jim Fulton, MIT X Consortium
 */

#ifndef _SSMailboxP_h
#define _SSMailboxP_h

#include "SSMailbox.h"
#include "regexp.h"


#ifdef SYSV
#define MAILBOX_DIRECTORY "/usr/mail"
#else
#define MAILBOX_DIRECTORY "/usr/spool/mail"
#endif

typedef struct bindings_struct
{
    regexp *pattern;
    String sound;
    String face;
    struct bindings_struct *next;
} Bindings;

typedef struct {			/* new fields for mailbox widget */
    /* resources */
    int update;				/* seconds between updates */
    Pixel foreground_pixel;		/* color index of normal state fg */
    String filename;			/* filename to watch */
    String check_command;		/* command to exec for mail check */
    Boolean reverseVideo;		/* do reverse video? */
    Boolean flipit;			/* do flip of full pixmap */
    int volume;				/* bell volume */
    Boolean once_only;			/* ring bell only once on new mail */
    /* local state */
    GC gc;				/* normal GC to use */
    long last_numnew;			/* last number of new msgs */
    XtIntervalId interval_id;		/* time between checks */
    Boolean flag_up;			/* is the flag up? */
    struct _mbimage {
	Pixmap bitmap, mask;		/* depth 1, describing shape */
	XImage *image;			/* full depth XImage */
	XColor *colors;			/* the colors in this image */
	int ncolors;			/* this is 0 for monochrome xbms */
	int width, height;		/* geometry of pixmaps */
    } full, empty, current_face;
#ifdef SHAPE
    Boolean shapeit;			/* do shape extension */
    struct {
	Pixmap mask;
	int x, y;
    } shape_cache;			/* last set of info */
#endif
    String sound;                     /* name of default sound to play */
    String no_mail_sound;             /* sound to play when no mail */
    String face;                      /* name of the bitmap to show */
    String sound_bindings;	      /* bindings of sounds to sender names */
    String face_bindings;	      /* bindings of bitmaps to sender names */
    Bindings *bindings;		      /* sounds and faces */
    Boolean case_sensitive;
    Boolean auto_resize;
    Boolean not_on_console;

    Boolean use_LED;		/* whether to use a kbd LED for reporting */
    int which_LED;		/* if so, which one */
} SSMailboxPart;

typedef struct _SSMailboxRec {		/* full instance record */
    CorePart core;
    SSMailboxPart ssmailbox;
} SSMailboxRec;


typedef struct {			/* new fields for mailbox class */
    int dummy;				/* stupid C compiler */
} SSMailboxClassPart;

typedef struct _SSMailboxClassRec {	/* full class record declaration */
    CoreClassPart core_class;
    SSMailboxClassPart ssmailbox_class;
} SSMailboxClassRec;

extern SSMailboxClassRec ssmailboxClassRec;	 /* class pointer */

#endif /* _SSMailboxP_h */
